package
{
import flash.display.Stage;
import flash.events.KeyboardEvent;
import flash.geom.Vector3D;
import flash.geom.Matrix3D;
import flash.ui.Keyboard;

/*
** A 3D camera using perspective projection.
** @author Jackson Dunstan
**
** To which we add the velocity/acceleration controls from
** Mario Scabia's example.
*/
public class DCamera
{
	private const MIN_NEAR_DISTANCE:Number = 0.001;
	private const MIN_PLANE_SEPARATION:Number = 0.001;
	private const EPSILON:Number = 0.001;
	private const MAX_LINEAL_VELOCITY:Number = 0.05;
	private const MAX_ROTATIONAL_VELOCITY:Number = 0.5;
	private const LINEAR_ACCELERATION:Number = 0.0005;
	private const ROTATIONAL_ACCELERATION:Number = 0.01;
	private const DAMPING:Number = 1.09;

	private var __position:Vector3D;
	private var __target:Vector3D;
	private var __upDir:Vector3D;
	private var __realUpDir:Vector3D;
	private var __rightDir:Vector3D;
	private var __viewDir:Vector3D;
	private var __viewDirMag:Number;
	private var __near:Number;
	private var __far:Number;
	private var __aspect:Number;
	private var __vFOV:Number;
	private var __worldToView:Matrix3D;
	private var __viewToClip:Matrix3D;
	private var __worldToClip:Matrix3D;
	private var __tempWorldToViewMatrix:Matrix3D;

	private var zoomAcceleration:Number;
	private var zoomVelocity:Number;
	private var horizontalAcceleration:Number;
	private var horizontalVelocity:Number;
	private var verticalAcceleration:Number;
	private var verticalVelocity:Number;
	private var pitchAcceleration:Number;
	private var pitchVelocity:Number;
	private var yawAcceleration:Number;
	private var yawVelocity:Number;

	/**
	*   Make the camera
	*   @param near Distance to the near clipping plane. Capped to MIN_NEAR_DISTANCE.
	*   @param far Distance to the far clipping plane. Must be MIN_PLANE_SEPARATION greater than near.
	*   @param aspect Aspect ratio of the camera lens
	*   @param vFOV Vertical field of view
	*   @param positionX X component of the camera's position
	*   @param positionY Y component of the camera's position
	*   @param positionZ Z component of the camera's position
	*   @param targetX X compoennt of the point the camera is aiming at
	*   @param targetY Y compoennt of the point the camera is aiming at
	*   @param targetZ Z compoennt of the point the camera is aiming at
	*   @param upDirX X component of the direction considered to be "up"
	*   @param upDirX X component of the direction considered to be "up"
	*   @param upDirY Y component of the direction considered to be "up"
	*/
	public function
	DCamera (stage:Stage,
		near:Number, far:Number, aspect:Number, vFOV:Number,
		positionX:Number, positionY:Number, positionZ:Number,
		targetX:Number, targetY:Number, targetZ:Number,
		upDirX:Number, upDirY:Number, upDirZ:Number)
	{
		if (near < MIN_NEAR_DISTANCE) near = MIN_NEAR_DISTANCE;
		if (far < near+MIN_PLANE_SEPARATION) far = near + MIN_PLANE_SEPARATION;

		__near = near;
		__far = far;
		__aspect = aspect;
		__vFOV = vFOV;
		__position = new Vector3D (positionX, positionY, positionZ);
		__target = new Vector3D (targetX, targetY, targetZ);
		__upDir = new Vector3D (upDirX, upDirY, upDirZ);
		__upDir.normalize();

		__viewDir = new Vector3D();
		__rightDir = new Vector3D();
		__realUpDir = new Vector3D();
		__tempWorldToViewMatrix = new Matrix3D();

		__worldToView = new Matrix3D();
		__viewToClip = new Matrix3D();
		__worldToClip = new Matrix3D();

		updateWorldToView();
		updateViewToClip();
		updateWorldToClip();

		zoomAcceleration = horizontalAcceleration = verticalAcceleration = 0.0;
		yawAcceleration = pitchAcceleration = 0.0;
		zoomVelocity = horizontalVelocity = verticalVelocity = 0.0;
		yawVelocity = pitchVelocity = 0.0;

		stage.addEventListener (KeyboardEvent.KEY_DOWN, keyDownHandler);
		stage.addEventListener (KeyboardEvent.KEY_UP, keyUpHandler);
	} // End of Constructor for DCamera.

	public function
	get worldToClipMatrix()
	:Matrix3D
	{
		yawVelocity = calculateUpdatedVelocity (yawVelocity, yawAcceleration, MAX_ROTATIONAL_VELOCITY);
		if (yawVelocity) rotate(yawVelocity, __realUpDir);

		pitchVelocity = calculateUpdatedVelocity (pitchVelocity, pitchAcceleration, MAX_ROTATIONAL_VELOCITY);
		if (pitchVelocity) rotate(pitchVelocity, __rightDir);

		zoomVelocity = calculateUpdatedVelocity (zoomVelocity, zoomAcceleration, MAX_LINEAL_VELOCITY);
		if (zoomVelocity) moveAlongAxis (zoomVelocity, __viewDir);

		horizontalVelocity = calculateUpdatedVelocity (horizontalVelocity, horizontalAcceleration, MAX_LINEAL_VELOCITY);
		if (horizontalVelocity) moveAlongAxis (horizontalVelocity, __rightDir);

		verticalVelocity = calculateUpdatedVelocity (verticalVelocity, verticalAcceleration, MAX_LINEAL_VELOCITY);
		if (verticalVelocity) moveAlongAxis (verticalVelocity, __upDir);

		return (__worldToClip);
	} // End of worldToClipMatrix getter.

	private function
	calculateUpdatedVelocity (velocity:Number, acceleration:Number, maxVelocity:Number)
	:Number
	{
		var newVelocity:Number;
		if (acceleration != 0) {
			newVelocity = (velocity + acceleration);
			if (newVelocity > maxVelocity) {
				newVelocity = maxVelocity;
			} else if (newVelocity < -maxVelocity) {
				newVelocity = -maxVelocity;
			}
		} else {
			newVelocity = (velocity / DAMPING);
			if (Math.abs (newVelocity) < EPSILON) newVelocity = 0.0;
		}
		return (newVelocity);
	} // End of calculateUpdatedVelocity().

	public function get positionX():Number { return (__position.x); }
	public function set positionX(x:Number): void
	{
		__position.x = x;
		updateWorldToView();
		updateWorldToClip();
	}

	public function get positionY(): Number { return (__position.y); }
	public function set positionY(y:Number): void
	{
		__position.y = y;
		updateWorldToView();
		updateWorldToClip();
	}

	public function get positionZ(): Number { return (__position.z); }
	public function set positionZ(z:Number): void
	{
		__position.z = z;
		updateWorldToView();
		updateWorldToClip();
	}

	public function setPositionValues (x:Number, y:Number, z:Number): void
	{
		__position.x = x;
		__position.y = y;
		__position.z = z;
		updateWorldToView();
		updateWorldToClip();
	}

	public function get targetX(): Number { return (__target.x); }
	public function set targetX(x:Number): void
	{
		__target.x = x;
		updateWorldToView();
		updateWorldToClip();
	}

	public function get targetY(): Number { return (__target.y); }
	public function set targetY(y:Number): void
	{
		__target.y = y;
		updateWorldToView();
		updateWorldToClip();
	}

	public function get targetZ(): Number { return (__target.z); }
	public function set targetZ(z:Number): void
	{
		__target.z = z;
		updateWorldToView();
		updateWorldToClip();
	}

	public function setTargetValues (x:Number, y:Number, z:Number): void
	{
		__target.x = x;
		__target.y = y;
		__target.z = z;
		updateWorldToView();
		updateWorldToClip();
	}

	public function get near(): Number { return (__near); }
	public function set near(near:Number): void
	{
		__near = near;
		updateViewToClip();
		updateWorldToClip();
	}

	public function get far(): Number { return (__far); }
	public function set far(far:Number): void
	{
		__far = far;
		updateViewToClip();
		updateWorldToClip();
	}

	public function get vFOV(): Number { return (__vFOV); }
	public function set vFOV(vFOV:Number): void
	{
		__vFOV = vFOV;
		updateViewToClip();
		updateWorldToClip();
	}

	public function get aspect(): Number { return (__aspect); }
	public function set aspect(aspect:Number): void
	{
		__aspect = aspect;
		updateViewToClip();
		updateWorldToClip();
	}

	private function moveAlongAxis (units:Number, axis:Vector3D): void
	{
		var delta:Vector3D = axis.clone();
		delta.scaleBy (units);

		var newPos:Vector3D = __position.add(delta);
		setPositionValues (newPos.x, newPos.y, newPos.z);

		var newTarget:Vector3D = __target.add(delta);
		setTargetValues (newTarget.x, newTarget.y, newTarget.z);
	}

	private function rotate (numDegrees:Number, axis:Vector3D): void
	{
		if (isNaN (numDegrees)) return;

		// Make positive and negative make sense
		numDegrees = -numDegrees;

		var rotMat:Matrix3D = new Matrix3D();
		rotMat.appendRotation (numDegrees, axis);

		var rotatedViewDir:Vector3D = rotMat.transformVector (__viewDir);
		rotatedViewDir.scaleBy (__viewDirMag);

		var newTarget:Vector3D = __position.add(rotatedViewDir);

		setTargetValues (newTarget.x, newTarget.y, newTarget.z);
	}

	private function updateWorldToView(): void
	{
		// viewDir = target - position
		var viewDir:Vector3D = __viewDir;
		viewDir.x = __target.x - __position.x;
		viewDir.y = __target.y - __position.y;
		viewDir.z = __target.z - __position.z;
		__viewDirMag = __viewDir.normalize();

		// Up is already normalized
		var upDir:Vector3D = __upDir;

		// rightDir = viewDir X upPrime
		var rightDir:Vector3D = __rightDir;
		rightDir.x = viewDir.y*upDir.z - viewDir.z*upDir.y;
		rightDir.y = viewDir.z*upDir.x - viewDir.x*upDir.z;
		rightDir.z = viewDir.x*upDir.y - viewDir.y*upDir.x;

		// realUpDir = rightDir X viewDir
		var realUpDir:Vector3D = __realUpDir;
		realUpDir.x = rightDir.y*viewDir.z - rightDir.z*viewDir.y;
		realUpDir.y = rightDir.z*viewDir.x - rightDir.x*viewDir.z;
		realUpDir.z = rightDir.x * viewDir.y - rightDir.y * viewDir.x;
		/*
		trace (
			"\n_ViewDir:", __viewDir,
			"\nRealDir:", realUpDir,
			"\nRightDir:", rightDir,
			"\nViewDir:", viewDir,
			"\nTarget:", __target);
		*/

		// Translation by -position
		var rawData:Vector.<Number> = __worldToView.rawData;
		rawData[0] = 1;
		rawData[1] = 0;
		rawData[2] = 0;
		rawData[3] = -__position.x;
		rawData[4] = 0;
		rawData[5] = 1;
		rawData[6] = 0;
		rawData[7] = -__position.y;
		rawData[8] = 0;
		rawData[9] = 0;
		rawData[10] = 1;
		rawData[11] = -__position.z;
		rawData[12] = 0;
		rawData[13] = 0;
		rawData[14] = 0;
		rawData[15] = 1;
		__worldToView.rawData = rawData;

		// Look At matrix. Some parts of this are constant.
		rawData = __tempWorldToViewMatrix.rawData;
		rawData[0] = rightDir.x;
		rawData[1] = rightDir.y;
		rawData[2] = rightDir.z;
		rawData[3] = 0;
		rawData[4] = realUpDir.x;
		rawData[5] = realUpDir.y;
		rawData[6] = realUpDir.z;
		rawData[7] = 0;
		rawData[8] = -viewDir.x;
		rawData[9] = -viewDir.y;
		rawData[10] = -viewDir.z;
		rawData[11] = 0;
		rawData[12] = 0;
		rawData[13] = 0;
		rawData[14] = 0;
		rawData[15] = 1;
		__tempWorldToViewMatrix.rawData = rawData;

		__worldToView.prepend(__tempWorldToViewMatrix);
	}

	private function updateViewToClip(): void
	{
		var f:Number = 1.0 / Math.tan(__vFOV);
		__viewToClip.rawData = new <Number>[
			(f / __aspect), 0, 0, 0,
			0, f, 0, 0,
			0, 0, ((__far+__near)/(__near-__far)), ((2*__far*__near)/(__near-__far)),
			0, 0, -1, 0
			];
	}

	private function updateWorldToClip(): void
	{
		__worldToView.copyToMatrix3D(__worldToClip);
		__worldToClip.prepend(__viewToClip);
	}

	private function
	keyDownHandler (event:KeyboardEvent)
	:void
	{
		switch (event.keyCode) {
			case (Keyboard.UP):
				if (event.shiftKey) {
					pitchAcceleration = ROTATIONAL_ACCELERATION;
				} else {
					verticalAcceleration = LINEAR_ACCELERATION;
				}
				break;
			case (Keyboard.DOWN):
				if (event.shiftKey) {
					pitchAcceleration = -ROTATIONAL_ACCELERATION;
				} else {
					verticalAcceleration = -LINEAR_ACCELERATION;
				}
				break;
			case (Keyboard.LEFT):
				if (event.shiftKey) {
					yawAcceleration = -ROTATIONAL_ACCELERATION;
				} else {
					horizontalAcceleration = -LINEAR_ACCELERATION;
				}
				break;
			case (Keyboard.RIGHT):
				if (event.shiftKey) {
					yawAcceleration = ROTATIONAL_ACCELERATION;
				} else {
					horizontalAcceleration = LINEAR_ACCELERATION;
				}
				break;
			case (Keyboard.HOME):
				zoomAcceleration = LINEAR_ACCELERATION;
				break;
			case (Keyboard.END):
				zoomAcceleration = -LINEAR_ACCELERATION;
				break;
			default:
				// trace ("Unexpected Key Code", event.keyCode);
				return;
		} // End of Switch on KeyCode.
	} // End of keyDownHandler().

	private function
	keyUpHandler (event:KeyboardEvent)
	:void
	{
		switch (event.keyCode) { 
			case (Keyboard.UP):
			case (Keyboard.DOWN):
				pitchAcceleration = verticalAcceleration = 0;
				break;
			case (Keyboard.LEFT):
			case (Keyboard.RIGHT):
				yawAcceleration = horizontalAcceleration = 0;
				break;
			case (Keyboard.HOME):
			case (Keyboard.END):
				zoomAcceleration = 0;
				break;
		} // End of Switch on KeyCode.
	} // End of keyUpHandler().

} // End of DCamera Class.

} // End of Package Declaration.
